/* ll_modem.c : lowlevel modem handling */
/*
    Copyright (C) 1994 1995 1996 1997 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact1: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "tomsysdep.h"

#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>


#include "io.h"
#include "modem.h"
#include "log.h"
#include "prototyp.h"
#include "box_conf.h"


static void put_rschar (int fd, const char *text);
static RETSIGTYPE alarm_handler (int sig);

/*
 * Ausgabe einer Zeichenkette an die serielle Schnittstelle
 * 
 * 07.07.93 Frank: optimierungsbedrftig wg. aufwendigem mintlib-write
 * 14.02.94 Frank, optimiert, Zusammenfassung der Ausgaben
 */
static void 
put_rschar (int fd, const char *text)
{
	while (*text)
	{
		char *p;

		p = strchr (text, '~');
		if (p)
		{
			if (p != text)
				write (fd, text, p - text);
			text = p + 1;
			msleep (500);
		}
		else
		{
			write (fd, text, strlen (text));
			text += strlen (text);
		}
	}

}								/* put_rschar */




static BOOLEAN got_alarm = FALSE;
static RETSIGTYPE
alarm_handler (int sig TOM_ATTRIB_UNUSED)
{
	got_alarm = TRUE;
}

/*
 * Zeichenkette ans Modem senden und Echo verwerfen
 */
void 
send_modem (int fd, const char *befehl)
{
	BOOLEAN first = TRUE;

  retry:
	put_rschar (fd, befehl);
	msleep (100);
	if (box_conf.modem_sends_echo)
	{
		char dummy[80];

		RETSIGTYPE (*old) (int);
		got_alarm = FALSE;
		old = signal (SIGALRM, alarm_handler);
		alarm (1);
		read (fd, dummy, (unsigned) strlen (befehl));
		alarm (0);
		if (got_alarm)
		{
		   /* Hm? */
			syslog (LOG_WARNING, "kein Echo vom Modem");
			if (first)
			{
				first = FALSE;
				io_send_break (fd);
				msleep (500);
				goto retry;
			}
		}
		signal (SIGALRM, old);
	}
}								/* send_modem */



#if 0
/* franks version ohne alarm drumherum */
/*
	timeout_sec gilt fr jeweils 128 Zeichen
*/
char * 
modem_readline(int fd, ULONG timeout_sec, BOOLEAN *timed_out)
{
	static char readbuf[1024];
	static int zurueckgestellt=0;

	size_t place=0;
	int zeichen;
	size_t timeout_counter=0;
	char *start=NULL;
	char *ptr=NULL;
	time_t end_time;

	end_time = time(NULL) + timeout_sec;

	if (timed_out)
		*timed_out=FALSE;

	while (time(NULL) <= end_time)
	{
		char *p;

		if (zurueckgestellt)
		{
			zeichen=zurueckgestellt;
			zurueckgestellt=0;
		}
		else
		{
			zeichen=read(fd,readbuf,sizeof(readbuf));
			if (zeichen==0)
			{
				msleep(500);
				continue;
			}
		}
		if (zeichen < 0)
			break;

		timeout_counter += zeichen;
		if (timeout_counter>128)
		{
			timeout_counter=0;
			end_time = time(NULL) + timeout_sec;
		}
		if (!start || (ptr+zeichen+1>=start+place))
		{
			size_t off;
			if (start)
				off=ptr-start;
			else
				off=0;
			if (zeichen>1024)
				place+=zeichen*4;
			else
				place+=1024;
			if (!start)
				start=malloc(place);
			else
				start=realloc(start,place);
			if (!start)
			{
				errno=ENOMEM;
				return NULL;
			}
			ptr=start+off;
		}
		memcpy(ptr,readbuf,zeichen);
		ptr[zeichen]=0;

		p=ptr;
		ptr+=zeichen;
		while (zeichen && *p!='\r' && *p!='\n')
		{
			p++;
			zeichen--;
		}
		if (*p=='\r' || *p=='\n')
		{
			if (*p=='\r' && zeichen)
			{
				*p=0;
				p++;
				zeichen--;
			}
			if (*p=='\n' && zeichen)
			{
				*p=0;
				p++;
				zeichen--;
			}
			if (zeichen)
			{
				memcpy(readbuf,p,zeichen);
				zurueckgestellt=zeichen;
			}
			*p=0;
			return start;
		}
	}

	if (start)
		free(start);
	if (timed_out)
		*timed_out=TRUE;

	return NULL;

} /* modem_readline */

#else

char * 
modem_readline(int fd, ULONG timeout_sec, BOOLEAN *timed_out)
{
	static char readbuf[1024];
	static size_t zurueckgestellt=0;

	size_t place=0;
	int zeichen;
	size_t timeout_counter=0;
	char *start=NULL;
	char *ptr=NULL;

	signal(SIGALRM,alarm_handler);
	alarm(timeout_sec);
	got_alarm=FALSE;
	if (timed_out)
		*timed_out=FALSE;

	while (TRUE)
	{
		char *p;
		if (got_alarm)
		{
			if (start)
				free(start);
			if (timed_out)
				*timed_out=TRUE;
			return NULL;
		}

		if (zurueckgestellt)
		{
			zeichen=zurueckgestellt;
			zurueckgestellt=0;
		}
		else
		{
			zeichen=read(fd,readbuf,sizeof(readbuf));
			if (zeichen==0)
			{
				msleep(500);
				continue;
			}
			if (got_alarm)
			{
				if (start)
					free(start);
				if (timed_out)
					*timed_out=TRUE;
				return NULL;
			}
		}
		if (zeichen<0)
		{
			if (start)
				free(start);
			return NULL;
		}
		timeout_counter+=zeichen;
		if (timeout_counter>128)
		{
			timeout_counter=0;
			alarm(timeout_sec);
		}
		if (!start || (ptr+zeichen+1>=start+place))
		{
			size_t off;
			if (start)
				off=ptr-start;
			else
				off=0;
			if (zeichen>1024)
				place+=zeichen*4;
			else
				place+=1024;
			if (!start)
				start=malloc(place);
			else
				start=realloc(start,place);
			if (!start)
			{
				errno=ENOMEM;
				return NULL;
			}
			ptr=start+off;
		}
		memcpy(ptr,readbuf,zeichen);
		ptr[zeichen]=0;

		p=ptr;
		ptr+=zeichen;
		while (zeichen && *p!='\r' && *p!='\n')
		{
			p++;
			zeichen--;
		}
		if (*p=='\r' || *p=='\n')
		{
			alarm(0);
			if (*p=='\r' && zeichen)
			{
				*p=0;
				p++;
				zeichen--;
			}
			if (*p=='\n' && zeichen)
			{
				*p=0;
				p++;
				zeichen--;
			}
			if (zeichen)
			{
				memcpy(readbuf,p,zeichen);
				zurueckgestellt=zeichen;
			}
			*p=0;
			return start;
		}
	}
	alarm(0);
	return start;
} /* modem_readline */

#endif
