/* mmread.h : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#ifndef MM_READ_H
#define MM_READ_H

/*
 * Kleiner Hack zum beschleunigten Lesen von Dateien. Benutzung:
 *    mm_file_t *in;
 *    char *buffer;
 *    size_t buffersize;
 *    size_t got;
 *
 *    in = MM_READ_OPEN(fname_in);
 *    if (!(MM_READ_IS_MMAPPED(in)))
 *    {
 *        buffersize=1024;
 *        buffer=XMALLOC(buffersize);
 *    }
 *    while ((got=MM_READ_GETLINE(in,&buffer,&buffersize))!=-1)
 *        do_something_with_buffer_but_dont_change_it();
 *    if (!(MM_READ_IS_MMAPPED(in)))
 *        free(buffer);
 *    MM_READ_CLOSE(in);
 *
 * Zustzliches Makro:
 *    MM_READ_SIZE(in) -> Lnge der Datei in Bytes.
 *
 * Wenn mglich wird mmap verwendet. Falls das nicht funktioniert wird
 * GNU getdelim verwendet (erspart einen strlen-Aufruf), sonst die
 * Kombimation von fgets und strlen.
 * 
 * der MM_READ_GETLINE bergebene bufferpointer wird verndert - 
 * mglicherweise durch getdelim (das realloc benutzen kann!) oder
 * bei erfolgreichem mmap() auf den Beginn der nchsten Zeile 
 * umgebogen.
 *
 * Achtung: nach MM_READ_GETLINE ist der bergebene Speicherbereich 
 * als READONLY zu betrachten!
 */

#ifndef HAVE_GETDELIM
#define getdelim(bufp,buflenp,delim,file) \
	((fgets(*(bufp),*(buflenp),file)==NULL) ? -1 : strlen(*(bufp)))
#endif

#ifdef HAVE_MMAP
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#endif
#include <sys/stat.h>

typedef struct
{
   /* private */
	size_t size;				/* size of file in bytes */
	FILE *file;
	int is_mmapped;
#ifdef HAVE_MMAP
	caddr_t start_addr;
	caddr_t read_addr;
	size_t to_read;
#endif
} mm_file_t;

static
#ifdef __GNUC__
  __inline__
#endif
  mm_file_t *
mm_read_open (const char *fname)
{
#ifdef HAVE_MMAP
	int fd;

#endif
	struct stat st;
	mm_file_t *f = malloc (sizeof (mm_file_t));

	if (!f)
	{
		fprintf (stderr, "out of memory\r\n");
		return NULL;
	}
#ifdef HAVE_MMAP
	fd = open (fname, O_RDONLY);
	if (fd < 0)
	{
		fprintf (stderr, "cannot open %s: %s\r\n", fname, strerror (errno));
		free (f);
		return NULL;
	}
	if (fstat (fd, &st) == -1)
	{
		fprintf (stderr, "cannot fstat %s: %s\r\n", fname, strerror (errno));
		free (f);
		close (fd);
		return NULL;
	}
	f->size=st.st_size;
	if (st.st_size)
		f->start_addr = mmap (0, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (st.st_size && f->start_addr)
	{
		f->is_mmapped = 1;
		f->to_read = st.st_size;
		f->read_addr = f->start_addr;
		close (fd);
	}
	else
	{
		f->is_mmapped = 0;
		f->file = fdopen (fd, "rb");
		if (!f->file)
		{
			fprintf (stderr, "cannot fdopen %s: %s\r\n", fname, strerror (errno));
			close (fd);
			free (f);
			return NULL;
		}
	}
#else
	f->file = fopen (fname, "rb");
	if (!f->file)
	{
		fprintf (stderr, "cannot fdopen %s: %s\r\n", fname, strerror (errno));
		free (f);
		return NULL;
	}
	if (fstat (fileno (f->file), &st) != 0)
	{
		fprintf (stderr, "cannot fstat %s: %s\r\n", fname, strerror (errno));
		fclose (f->file);
		free (f);
		return NULL;
	}
	f->size = st.st_size;
#endif
	return f;
}
#define MM_READ_OPEN(fname) mm_read_open((fname))

#ifdef HAVE_MMAP
static 
#ifdef __GNUC__
  __inline__
#endif
  size_t
mm_getline (mm_file_t * f, char **inbuf)
{
	char *p;
	size_t ret;

	if (!f->to_read)
	{
		*inbuf = NULL;
		return -1;
	}
	p = memchr (f->read_addr, '\n', f->to_read);
	*inbuf = f->read_addr;
	if (!p)
	{
		ret = f->to_read;
		f->read_addr = NULL;
	}
	else
	{
		ret = p - f->read_addr + 1;
		f->read_addr = p + 1;
	}
	f->to_read -= ret;
	return ret;
}
#define MM_READ_GETLINE(f,bufp,buflenp) \
	(!(f->is_mmapped) ? (getdelim((bufp),(buflenp),'\n',f->file)) \
		: (mm_getline (f, bufp)))
#define MM_READ_CLOSE(f) \
	if (f->is_mmapped) munmap (f->start_addr, f->size); \
	else if (f->file) fclose(f->file); \
	free(f); f=NULL;

#else

#define MM_READ_GETLINE(f,bufp,buflenp) \
	(getdelim((bufp),(buflenp),'\n',f->file))
#define MM_READ_CLOSE(f) \
	if (f && f->file) fclose(f->file); \
	if (f) {free(f); f=NULL; }
#endif

#define MM_READ_IS_MMAPPED(f) (f->is_mmapped)
#define MM_READ_SIZE(f) (f->size)

#endif
