/*
    Copyright (C) 1995 1996 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
*/
#include "tomsysdep.h"

#include "portab.h"
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>

#include "prototyp.h"
#include "modem.h"
#include "user.h"
#include "tompaths.h"
#include "routing.h"
#include "box_conf.h"
#include "libquark.h"
#include "tausch.h" /* ta_do_log1 */
#include "route_int.h"

int r_verbose_grade=0;

void
r_writecomment(r_loggrade_t grade, r_adr_t *adr, ...)
{
	va_list ap;
	const char *s;
	char *neu=NULL;
	static char *eoom=NULL;

	/* dieser kleine Hack gibt uns eine Restchance auf einen Logstring ... */
	if (!eoom)
		eoom=strdup("out of memory");

	/* Wichtigere Meldungen nicht berschreiben */
	if (adr->loggrade>grade)
		return;

	va_start(ap,adr);
	s=va_arg(ap,char *);
	vasprintf(&neu,s,ap);
	va_end(ap);
	if (!neu)
	{
		adr->logstr=eoom;
		eoom=NULL;
		return;
	}
	else
		adr->logstr=neu;

	adr->loggrade=grade;
}

void
r_printf(const char *format, ...)
{
	char tmp[1024];
	va_list param;

	va_start(param,format);
	vsprintf(tmp,format,param);
	va_end(param);

	mboutput(tmp);
}

char *
r_get_dsn(const char *format, ...)
{
	char *neu=NULL;
	va_list param;

	if (format) {
		va_start(param,format);
		vasprintf(&neu,format,param);
		va_end(param);
		if (neu)
			return neu;
	}
	return r_dsn_out_of_memory;
}

void
r_clean(r_adr_t *adr, BOOLEAN force)
{
#define FREESTR(str) do { \
	if (force) { \
		adr->str=NULL; \
	} else if (adr->str) \
	{ \
		free(adr->str); \
		adr->str=NULL; \
	} } while (0);

	FREESTR(erroradr);
	FREESTR(inputadr);
	FREESTR(resultadr);
	FREESTR(username);

	FREESTR(logstr);
	adr->loggrade=R_LOG_INFO;
	adr->error=FALSE;
	adr->errorflags=0;
	if (adr->dsn!=r_dsn_out_of_memory)
		FREESTR(dsn);

	adr->loopcount=0;

	adr->bezahlt=FALSE;
	adr->sec_bezahlt=FALSE;
	adr->level=FALSE;
	adr->sec_level=FALSE;
	adr->owner=0;

	adr->flags=0;

	adr->transport=NULL;
	adr->bezug=NULL;
	adr->thint=NULL;
	adr->user_gate_nr=0;

	FREESTR(infotext);

	if (force)
		adr->next=NULL;
}

r_adr_t *
r_alloc(void)
{
	r_adr_t *n;

	n=malloc(sizeof(r_adr_t));
	if (!n)
		return NULL;
	r_clean(n,TRUE);
	return n;
}

void 
r_free(r_adr_t *adr)
{
	r_clean(adr,FALSE);
	adr->next=NULL;
	free(adr);
}

void 
r_free_chain(r_adr_t *adr)
{
	r_adr_t *this;
	while(adr)
	{
		this=adr;
		adr=adr->next;
		r_free(this);
	}
}

void 
r_paniclog (const char *format,...)
{
	FILE *f;
	va_list param;
	char fname[PATH_MAX];
	time_t t=time(NULL);

	va_start (param, format);

	sprintf (fname, "%s%s.panic", box_conf.log_dir, datum_pro_S());
	f = fopen (fname, "a");
	if (f)
	{
		fprintf(f,"%s",ctime(&t));
		vfprintf(f,format,param);
		fprintf(f,"\n");
		fclose(f);
	}
	vfprintf(stderr,format,param);
	va_end (param);
}


/*
 * Diese Funktion ist nur ein Wrapper. Die eigentliche Arbeit geschieht
 * in ta_do_log1 (wo auch Bildschirmausgaben gemacht werden).
 * Logik:
 * - ffentliche Nachrichten werden generell geloggt.
 * - Statusnachrichten und Mails, fr die R_F_TAUSCHLOG gesetzt ist,
 *   werden via ta_do_log1 geloggt.
 * - andere Statusmeldungen und Mails werden nur auf den Bildschirm
 *   ausgegeben (das ist Debugging, nicht mehr).
 */
void
r_tauschuserlog(r_adr_t *adr, const char *format, ...)
{
	va_list param;

	char buf[2048];
	va_start (param, format);
	vsprintf(buf, format,param);
	va_end (param);

	if (!adr)
		ta_do_log1(buf);
	else if (adr->flags & R_F_TAUSCHLOG)
		ta_do_log1(buf);
	else
		mboutput(buf);
}

int 
r_set_verbose_grade(int ng)
{
	int old=r_verbose_grade;
	r_verbose_grade=ng;
	return old;
}

/*
 * Standardform fr hostnamen erzeugen: F.maus.de -> fue.maus.de
 */
void
r_hcanonical(const char *from, char *to)
{
	entumlauten(from,to);
	strlwr(to);
}

int
r_hostnamecmp(const char *dom1, const char *dom2)
{
	char buf1[256];
	char buf2[256];
	entumlauten(dom1,buf1);
	entumlauten(dom2,buf2);
	return stricmp(buf1,buf2);
}


const char *
get_path_entry(USERNR usernr)
{
	r_gateway_t *d;
	if (!r_init_done)
		r_do_init();

	d=r_gateway_anker;
	while (d)
	{
		if (d->via==usernr && d->path_entry)
			return d->path_entry;
		d=d->next;
	}
	return FALSE;
}



